from amaranth import Module, Cat, Mux, ShapeLike, Signal, Assert, unsigned
from amaranth.build import Platform
from amaranth.lib import wiring, data
from amaranth.lib.wiring import In, Out
from amaranth.lib.memory import Memory, ReadPort, WritePort
from amaranth.utils import ceil_log2


class RGBLayout(data.StructLayout):
    def __init__(self, r_bits, g_bits, b_bits):
        super().__init__(
            {
                "red": unsigned(r_bits),
                "green": unsigned(g_bits),
                "blue": unsigned(b_bits),
            }
        )


Rgb666Layout = RGBLayout(6, 6, 6)

Rgb111Layout = RGBLayout(1, 1, 1)


class Hub75Stream(wiring.Signature):
    """A Hub75E Driver for a single string of panels."""

    def __init__(self):
        super().__init__(
            {
                "latch": Out(1),
                "oe": Out(1),
                "addr": Out(5),
                "rgb0": Out(3),
                "rgb1": Out(3),
                "display_clk": Out(1),
            }
        )


class Hub75Ctrl(wiring.Signature):
    """Hub75E control/addressing signals.
    This is separated because certain designs keep these common
    between all panel outputs (e.g Colorlight 5A-75b boards).

    """

    def __init__(self):
        super().__init__(
            {
                "latch": Out(1),
                "oe": Out(1),
                "addr": Out(5),
                "display_clk": Out(1),
            }
        )


class Hub75Data(wiring.Signature):
    """Data lines for HUB75 displays. When combined with Hub75Ctrl, this forms a complete HUB75 interface.
    These are kept separate as some devices have a single set of control signals.
    """

    def __init__(self):
        super().__init__(
            {
                "rgb0": Out(Rgb111Layout),
                "rgb1": Out(Rgb111Layout),
            }
        )


class SwapBuffer(wiring.Component):
    """A pair of BRAMs for holdling line data that are swapped between using an external signal.

    Contains one write port, and one read port. They are attached to separate BRAMs to allow for
    one to be read at the same time another is being updated.

    Parameters:
        depth: i32, depth of each memory buffer
        shape: ShapeLike, the underlying shape of the memory.

    Signals:
        write_port: WritePort.Signature(width, shape)
        read_port: ReadPort.Signature(width, shape)
        selector: In(1)

    """

    def __init__(self, shape: ShapeLike, depth: int):
        super().__init__(
            {
                "selector": In(1),
                "read_port": In(
                    ReadPort.Signature(addr_width=ceil_log2(depth), shape=shape)
                ),
                "write_port": In(
                    WritePort.Signature(addr_width=ceil_log2(depth), shape=shape)
                ),
            }
        )
        self.data_shape = shape
        self.depth = depth

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        m.submodules.bram0 = self.bram0 = Memory(
            shape=self.data_shape, depth=self.depth, init=[]
        )
        m.submodules.bram1 = self.bram1 = Memory(
            shape=self.data_shape, depth=self.depth, init=[]
        )

        rd0 = self.bram0.read_port()
        wr0 = self.bram0.write_port()
        rd1 = self.bram1.read_port()
        wr1 = self.bram1.write_port()

        m.d.comb += [
            # wr addres
            wr0.addr.eq(self.write_port.addr),
            wr1.addr.eq(self.write_port.addr),
            # write enables are based on selector
            wr0.en.eq(~self.selector),
            wr1.en.eq(self.selector),
            # connect write data. This is FINE because
            # there is one driver (the external writer)
            # and we en based on selector so the other one isn't active
            wr0.data.eq(self.write_port.data),
            wr1.data.eq(self.write_port.data),
            # connect rd address lines
            rd0.addr.eq(self.read_port.addr),
            rd1.addr.eq(self.read_port.addr),
            rd0.en.eq(~self.selector),
            rd1.en.eq(self.selector),
            # we do this because the read_data lines are driven, this prevents
            # double-driver situations even though we en using selector above
            self.read_port.data.eq(Mux(self.selector, rd1.data, rd0.data)),
        ]

        return m


class Hub75StringDriver(wiring.Component):
    """A data driver for Hub75 panels. This accesses the line memory and feeds out the data.
    It is controlled by a Hub75Coordinator to signal when it should run and what bit of the data
    it should send.
    """

    def __init__(self, panel_length: int = 128, *, src_loc_at=0):
        self.panel_length = panel_length
        super().__init__(
            {
                "bcm_select": In(3),
                "done": Out(1),
                "start": In(1),
                "bram_port": In(
                    ReadPort.Signature(
                        addr_width=ceil_log2(panel_length),
                        shape=data.ArrayLayout(Rgb666Layout, 2),
                    )
                ),
                "display_out": Out(Hub75Data()),
            },
            src_loc_at=src_loc_at,
        )

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        self._counter = counter = Signal(32)  # unused count is optimized out
        m.d.sync += counter.eq(counter + 1)

        ram_rgb0_slice = Cat(
            self.bram_port.data[0]["red"].bit_select(self.bcm_select, 1),
            self.bram_port.data[0]["blue"].bit_select(self.bcm_select, 1),
            self.bram_port.data[0]["green"].bit_select(self.bcm_select, 1),
        )
        ram_rgb1_slice = Cat(
            self.bram_port.data[1]["red"].bit_select(self.bcm_select, 1),
            self.bram_port.data[1]["blue"].bit_select(self.bcm_select, 1),
            self.bram_port.data[1]["green"].bit_select(self.bcm_select, 1),
        )
        m.d.comb += [
            self.display_out.rgb0.eq(ram_rgb0_slice),
            self.display_out.rgb1.eq(ram_rgb1_slice),
        ]
        m.d.sync += Assert(self.bcm_select < 6)

        pixnum = Signal(ceil_log2(self.panel_length), init=self.panel_length - 1)
        m.d.comb += self.bram_port.addr.eq(pixnum)

        with m.FSM():
            with m.State("init"):
                m.d.sync += [
                    self.done.eq(0),
                    counter.eq(0),
                    pixnum.eq(self.panel_length - 1),
                ]
                with m.If(self.start == 1):
                    m.d.sync += self.bram_port.en.eq(1)
                    m.next = "writerow"
            with m.State("writerow"):
                with m.If(counter[0] == 0):
                    # do nothing
                    pass

                with m.If((counter[0] == 1) & (pixnum != 0)):
                    m.d.sync += pixnum.eq(pixnum - 1)

                with m.Else():
                    m.next = "done"
            with m.State("done"):
                m.d.sync += [self.done.eq(1), self.bram_port.en.eq(0)]
                m.next = "init"

        return m


class Hub75Coordinator(wiring.Component):
    """A shared-control hub75 driver"""

    def __init__(self, n_strings=1):
        self.n_strings = n_strings
        super().__init__()

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        return m


class Hub75EDriver(wiring.Component):
    """An optimized driver for hub75 strings.
    This version is faster than most implementations by merging the exposure
    period and the data-write period to happen simultaneously. As a result,
    the display can be brighter due to higher duty cycle.


    NOTICE: this is a direct port of the old verilog code. It isn't up to date with the
    modified structure. Notably, it uses RGB888 with double-fetch (4xclocking)
    """

    start: In(1)
    done: Out(1)
    out: Out(Hub75Stream())
    buf_addr: Out(9)
    buf_data: In(36)

    row_depth = 128
    bit_depth = 8
    bcm_len = 32

    def elaborate(self, platform: Platform) -> Module:
        m = Module()

        counter = Signal(32)
        bcm_shift: Signal = Signal(4, init=7)
        m.d.sync += counter.eq(counter + 1)
        # construct helper signals.
        ram_r = self.buf_data[16:23]
        ram_b = self.buf_data[8:15]
        ram_g = self.buf_data[0:7]

        ram_rgb_slice = Cat(
            ram_r.bit_select(bcm_shift, 1),
            ram_g.bit_select(bcm_shift, 1),
            ram_b.bit_select(bcm_shift, 1),
        )
        pixnum = Signal(8, reset=127)
        pixrow = Signal(1, reset=0)
        m.d.comb += self.buf_addr.eq(Cat(pixrow, pixnum))

        should_clock = counter < (128 * 2 + 1)
        should_expose = (counter < (32 << (bcm_shift + 1) + 1)).bool() & (
            bcm_shift != 7
        ).bool()

        with m.FSM():
            with m.State("init"):
                m.d.sync += [
                    bcm_shift.eq(7),
                    counter.eq(0),
                    self.done.eq(0),
                    pixnum.eq(127),
                    pixrow.eq(0),
                ]
                with m.If(self.start == 1):
                    m.next = "prefetch"
            with m.State("prefetch"):
                with m.If(counter == 0):
                    m.d.sync += pixrow.eq(0)
                with m.Elif(counter == 1):
                    m.d.sync += pixrow.eq(1)
                with m.Elif(counter == 2):
                    m.d.sync += self.out.rgb0.eq(ram_rgb_slice)
                with m.Elif(counter == 3):
                    m.d.sync += [self.out.rgb1.eq(ram_rgb_slice), counter.eq(0)]
                    m.next = "writerow"
            with m.State("writerow"):
                # expose if we haven't done it for long enough yet.
                m.d.sync += self.out.oe.eq(~(should_expose))
                with m.If(should_clock):
                    # clock is high on entry
                    m.d.sync += self.out.display_clk.eq(counter[1] == 0)

                    with m.If(counter[0:1] == 0):
                        # rising edge of the clock
                        m.d.sync += pixrow.eq(0)

                    with m.If(counter[0:1] == 1):
                        m.d.sync += [
                            pixrow.eq(1),
                            self.out.rgb0.eq(ram_rgb_slice),
                        ]
                    with m.If(counter[0:1] == 2):
                        m.d.sync += [
                            pixnum.eq(pixnum - 1),
                            pixrow.eq(0),
                            self.out.rgb1.eq(ram_rgb_slice),
                        ]

                with m.Elif(~(should_expose)):
                    # we are done both feeding in the new data and exposing the previous.
                    m.d.sync += [counter.eq(0), self.out.display_clk.eq(0)]
                    m.next = "latchout"
            with m.State("latchout"):
                m.d.sync += [
                    pixnum.eq(127),
                    self.out.latch.eq(1),
                ]
                with m.If(counter > 3):
                    m.d.sync += self.out.latch.eq(0)
                    m.d.sync += counter.eq(0)
                    with m.If(bcm_shift == 0):
                        m.next = "finish"
                    with m.Else():
                        m.d.sync += bcm_shift.eq(bcm_shift - 1)
                        m.next = "prefetch"

            with m.State("finish"):
                m.d.sync += Assert(bcm_shift == 0, "finish without bcm shift 0")

                with m.If(counter < (32)):
                    m.d.sync += self.out.oe.eq(0)
                with m.Else():
                    m.d.sync += [self.out.oe.eq(1), self.done.eq(1)]
                    m.next = "init"

        return m


if __name__ == "__main__":
    m = Hub75EDriver()
    from amaranth.cli import main

    main(m)
